"""Usage:
  tenzir-platform node list
  tenzir-platform node ping <node_id>
  tenzir-platform node create [--name=<node_name>]
  tenzir-platform node delete <node_id>
  tenzir-platform node run [--name=<node_name>] [--image=<container_image>]

Options:
  --name=<node_name>         The name of the newly created node [default: CLI_Node]
  --image=<container_image>  The docker image to use for the newly created node

Description:
  tenzir-platform node list
    Display a list of all nodes in the current workspace.

  tenzir-platform node ping
    Send a ping request to a node and measure the response time.

  tenzir-platform node run
    Create a new temporary node and run it in a local 'docker compose' stack.
    A new node is created in the currently selected workspace.
    The node is deleted when the command is interrupted.
    Requires the 'docker compose' binary in the current PATH.
"""

from tenzir_platform.helpers.cache import load_current_workspace
from tenzir_platform.helpers.client import AppClient
from tenzir_platform.helpers.environment import PlatformEnvironment
from pydantic import BaseModel
from docopt import docopt
from typing import Optional
import json
import time
import tempfile
import os
import subprocess
import re
import random


def list(client: AppClient, workspace_id: str):
    resp = client.post(
        "/list-nodes",
        json={
            "tenant_id": workspace_id,
        },
    )
    resp.raise_for_status()
    for i, node in enumerate(resp.json()["nodes"]):
        connection_symbol = "🟢" if node["lifecycle_state"] == "connected" else "🔴"
        print(f"{connection_symbol} {node['name']} ({node['node_id']})")


def ping(client: AppClient, workspace_id: str, node_id: str):
    start = time.time()
    resp = client.post(
        "proxy",
        json={
            "node_id": node_id,
            "tenant_id": workspace_id,
            "http": {
                "path": "/ping",
                "method": "POST",
            },
        },
    )
    if resp.status_code == 410:
        print("node disconnected")
        return
    elif resp.status_code != 200:
        print(f"error {resp.status_code} after {time.time()-start}s")
        return
    print(f"response {resp.status_code} in {time.time()-start}s")


def run(client: AppClient, workspace_id: str, node_name: str, container_image: Optional[str]):
    resp = client.post(
        "generate-client-config",
        json={
            "config_type": "docker",
            "tenant_id": workspace_id,
            "node_name": node_name,
        },
    )
    resp.raise_for_status()
    json = resp.json()
    node_id = json["node_id"]
    contents = json["contents"]

    # Hacky text replacement for using custom containers
    if container_image is not None:
        pattern = r"^(\s*image:) tenzir/tenzir.*$"
        replacement = r"\1 " + container_image
        contents = re.sub(pattern, replacement, contents, count=1, flags=re.MULTILINE)

    temp_file = tempfile.NamedTemporaryFile(delete=False, mode="w+")
    temp_file_name = temp_file.name
    temp_file.write(contents)
    temp_file.flush()

    try:
        print("running temporary Tenzir node")
        full_command = f"docker compose -f {temp_file_name} up"
        subprocess.run(full_command, shell=True, check=True)
    except KeyboardInterrupt:
        # Regular exit with CTRL-C
        pass
    except subprocess.CalledProcessError as e:
        print(f"Error running the docker compose command: {e}")
    finally:
        print("removing node and config file")
        if os.path.exists(temp_file_name):
            os.remove(temp_file_name)
        client.post(
            "delete-node",
            json={
                "tenant_id": workspace_id,
                "node_id": node_id,
            },
        )


def create(client: AppClient, workspace_id: str, node_name: Optional[str]):
    if node_name is None:
        time_suffix = datetime.datetime.now(datetime.UTC).strftime("%Y%m%dT%H%M%SZ")
        node_name = f"node-{time_suffix}"
    resp = client.post(
        "/generate-client-config",
        json={
            "tenant_id": workspace_id,
            "config_type": "docker",
            "node_name": node_name,
        },
    )
    resp.raise_for_status()
    print(f"created node {resp.json()['node_id']}")


def delete(client: AppClient, workspace_id: str, node_id: str):
    resp = client.post(
        "/delete-node",
        json={
            "tenant_id": workspace_id,
            "node_id": node_id,
        },
    )
    resp.raise_for_status()
    print(f"deleted node {node_id}")


def ping(client: AppClient, workspace_id: str, node_id: str):
    start = time.time()
    resp = client.post(
        "proxy",
        json={
            "node_id": node_id,
            "tenant_id": workspace_id,
            "http": {
                "path": "/ping",
                "method": "POST",
            },
        },
    )
    if resp.status_code == 410:
        print("node disconnected")
        return
    elif resp.status_code != 200:
        print(f"error {resp.status_code} after {time.time()-start}s")
        return
    print(f"response {resp.status_code} in {time.time()-start}s")

def node_subcommand(platform: PlatformEnvironment, argv):
    args = docopt(__doc__, argv=argv)
    try:
        workspace_id, user_key = load_current_workspace(platform)
        client = AppClient(platform=platform)
        client.workspace_login(user_key)
    except Exception as e:
        print(f"error: {e}")
        print(
            "Failed to load current workspace, please run 'tenzir-platform workspace select' first"
        )
        exit(1)

    if args["list"]:
        list(client, workspace_id)
    elif args["ping"]:
        node_id = args["<node_id>"]
        ping(client, workspace_id, node_id)
    elif args["run"]:
        node_name = args["--name"]
        container_image = args['--image']
        run(client, workspace_id, node_name, container_image)
    elif args["create"]:
        node_name = args["--name"]
        create(client, workspace_id, node_name, container_image)
    elif args["delete"]:
        node_id = args["<node_id>"]
        delete(client, workspace_id, node_id)




import hashlib
import os
import time

import jwt
import requests
from jwt import PyJWKClient
from typing import Optional, Any
from tenzir_platform.helpers.cache import filename_in_cache
from tenzir_platform.helpers.environment import PlatformEnvironment


class INVALID_API_KEY(Exception):
    pass


class ValidOidcToken:
    """The validated and decoded id_token.

    - Validate the presence of some claims required by the OIDC spec
    - Provide helper methods for checks"""

    def __init__(self, raw_oidc: dict[str, Any]) -> None:
        if "sub" not in raw_oidc or not isinstance(raw_oidc["sub"], str):
            raise INVALID_API_KEY("sub string required in OIDC token")
        self.user_id = raw_oidc["sub"]
        self._raw_oidc = raw_oidc

    def check_connection(self, connection: str) -> bool:
        return self.user_id.startswith(f"{connection}|")

    def get_claim_str(self, key: str, default: str) -> str:
        if key in self._raw_oidc and not isinstance(self._raw_oidc[key], str):
            raise INVALID_API_KEY(f"{key} is expected to be a string")
        return self._raw_oidc.get(key, default)

    @staticmethod
    def _is_list(val: Any) -> bool:
        return isinstance(val, list) and all(isinstance(x, str) for x in val)

    def get_claim_list(self, key: str, default: list[str]) -> list[str]:
        if key in self._raw_oidc and not ValidOidcToken._is_list(self._raw_oidc[key]):
            raise INVALID_API_KEY(f"{key} is expected to be a list of strings")
        return self._raw_oidc.get(key, default)

    def __str__(self) -> str:
        return self._raw_oidc.__str__()


class IdTokenClient:
    def __init__(self, platform: PlatformEnvironment):
        self.platform_environment = platform
        self.issuer = platform.oidc_issuer_url
        self.client_id = platform.oidc_client_id
        self.audience = platform.oidc_audience
        self.verbose = False

    def validate_token(self, id_token: str) -> ValidOidcToken:
        """Verify the token using the audience specific to the CLI"""
        jwks_url = f"{self.issuer.rstrip('/')}/.well-known/jwks.json"
        jwks_client = PyJWKClient(jwks_url)
        signing_key = jwks_client.get_signing_key_from_jwt(id_token)
        validated_token = jwt.decode(
            id_token,
            signing_key.key,
            algorithms=["RS256"],
            issuer=self.issuer,
            # for id tokens, the audience is the client_id
            audience=self.client_id,
        )
        return ValidOidcToken(validated_token)

    def reauthenticate_token(self) -> str:
        # TODO: Add an option to use password flow for non-interactive environments.
        token_data = self._device_code_flow()
        return self._unwrap_flow_result(token_data)

    def _device_code_flow(self) -> dict[str, str]:
        device_code_payload = {
            "client_id": self.client_id,
            "scope": "openid email",
            # This points to an API audience in Auth0
            "audience": self.audience,
        }
        device_code_response = requests.post(
            f"{self.issuer.rstrip('/')}/oauth/device/code",
            data=device_code_payload,
        )

        if device_code_response.status_code != 200:
            raise Exit(f"Error generating the device code: {device_code_response.text}")

        device_code_data = device_code_response.json()
        print(
            "1. On your computer or mobile device navigate to: ",
            device_code_data["verification_uri_complete"],
        )
        print(
            "2. Verify you're seeing the following code and confirm: ",
            device_code_data["user_code"],
        )
        print("3. Wait up to 10 seconds")

        token_payload = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": device_code_data["device_code"],
            "client_id": self.client_id,
        }
        authenticated = False
        while not authenticated:
            token_response = requests.post(
                f"{self.issuer.rstrip('/')}/oauth/token", data=token_payload
            )
            token_data = token_response.json()
            if token_response.status_code == 200:
                print("Authenticated!")
                break
            elif token_data["error"] not in ("authorization_pending", "slow_down"):
                print(token_data["error_description"])
                authenticated = True
            else:
                time.sleep(device_code_data["interval"])
        return token_data

    def _unwrap_flow_result(self, token_data: dict[str, str]) -> str:
        id_token = token_data["id_token"]
        current_user = self.validate_token(id_token)
        if self.verbose:
            print(f"obtained id_token: {current_user}")
        self._store_id_token(id_token)
        return id_token

    def _filename_in_cache(self):
        return filename_in_cache(self.platform_environment, "id_token")

    def _store_id_token(self, token: str):
        filename = self._filename_in_cache()
        print(f"saving token to {filename}")
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "w") as f:
            f.write(token)

    def load_id_token(self) -> str:
        filename = self._filename_in_cache()
        try:
            with open(filename, "r") as f:
                token = f.read()
            self.validate_token(token)
            return token
        except Exception:
            print("could not load valid token from cache, reauthenticating")
        return self.reauthenticate_token()



"""
Usage:
  tenzir-platform workspace list
  tenzir-platform workspace select <workspace_id>
"""

from tenzir_platform.helpers.client import AppClient, TargetApi
from tenzir_platform.helpers.environment import PlatformEnvironment
from tenzir_platform.helpers.cache import store_workspace
from tenzir_platform.helpers.oidc import IdTokenClient
from docopt import docopt
import json


def list(platform: PlatformEnvironment):
    """Get list of authorized workspaces for the current CLI user"""
    id_token = IdTokenClient(platform).load_id_token()
    app_cli = AppClient(platform=platform)
    resp = app_cli.post(
        "get-login-info",
        json={
            "id_token": id_token,
        },
        target_api=TargetApi.USER_PUBLIC,
    )
    resp.raise_for_status()
    for i, workspace in enumerate(resp.json()["allowed_tenants"]):
        print(f"{workspace['tenant_id']} - {workspace['name']}")


def select(platform: PlatformEnvironment, workspace_id: str):
    """Log in to a tenant as the current CLI user"""
    id_token = IdTokenClient(platform).load_id_token()
    app_cli = AppClient(platform)
    resp = app_cli.post(
        "switch-tenant",
        json={
            "id_token": id_token,
            "tenant_id": workspace_id,
        },
        target_api=TargetApi.USER_PUBLIC,
    )
    resp.raise_for_status()
    user_key = resp.json()["user_key"]
    store_workspace(platform, workspace_id, user_key)
    print(f"Switched to workspace {workspace_id}")


def workspace_subcommand(platform: PlatformEnvironment, argv):
    args = docopt(__doc__, argv=argv)
    if args["list"]:
        list(platform)
    if args["select"]:
        workspace_id = args["<workspace_id>"]
        select(platform, workspace_id)



from pydantic_settings import BaseSettings


API_ENDPOINT = "https://rest.tenzir.app/production-v1/"
OIDC_ISSUER_URL = "https://tenzir.eu.auth0.com/"
OIDC_CLIENT_ID = "vzRh8grIVu1bwutvZbbpBDCOvSzN8AXh"
OIDC_AUDIENCE = "tenzir_cli"


class PlatformEnvironment(BaseSettings):
    # The remote API endpoint of the platform.
    api_endpoint: str = API_ENDPOINT

    # An arbitrary short string to identify this environment
    # in the local cache directory.
    stage_identifier: str = "prod"

    # TODO: Provide a new `/oidc-config` endpoint in the public platform api
    # to load these values dynamically.
    oidc_issuer_url: str = OIDC_ISSUER_URL
    oidc_client_id: str = OIDC_CLIENT_ID
    oidc_audience: str = OIDC_AUDIENCE

    @staticmethod
    def load():
        return PlatformEnvironment(
            _env_prefix="TENZIR_PLATFORM_CLI", _env_nested_delimiter="__"
        )



"""Usage: tenzir-platform auth login

Login command.
"""

from tenzir_platform.helpers.oidc import IdTokenClient
from tenzir_platform.helpers.environment import PlatformEnvironment
from docopt import docopt


def login(platform: PlatformEnvironment):
    token_client = IdTokenClient(platform)
    token = token_client.load_id_token()
    decoded_token = token_client.validate_token(token)
    print(f"Logged in as {decoded_token.user_id}")


def auth_subcommand(platform: PlatformEnvironment, argv):
    args = docopt(__doc__, argv=argv)
    if args["login"]:
        login(platform)
